/**************************************************************************
*  Copyright (c) 2016 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  21.12.2016  mifi  First Version.
*                    Based on the original startup.s from the Altera
*                    Minimal Preloader (MPL) with rework and cleanup for 
*                    use with the GCC only.
*  28.12.2016  mifi  Added FPU support.
**************************************************************************/


/******************************************************************************
*
* Copyright 2014 Altera Corporation. All Rights Reserved.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
* 
* 1. Redistributions of source code must retain the above copyright notice,
* this list of conditions and the following disclaimer.
* 
* 2. Redistributions in binary form must reproduce the above copyright notice,
* this list of conditions and the following disclaimer in the documentation
* and/or other materials provided with the distribution.
* 
* 3. Neither the name of the copyright holder nor the names of its contributors
* may be used to endorse or promote products derived from this software without
* specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
* 
******************************************************************************/

/*=======================================================================*/
/*  Global symbols                                                       */
/*=======================================================================*/

   .global  _start

/*=======================================================================*/
/*  Internal Definitions                                                 */
/*=======================================================================*/

   /* 
    * Mode bit definitions, and interrupt flag masks 
    */
   .equ     Mode_USR,    0x10    /* Un-privileged mode */
   .equ     Mode_FIQ,    0x11    /* Entered as result of fast interrupt */
   .equ     Mode_IRQ,    0x12    /* Entered as result of normal interrupt */
   .equ     Mode_SVC,    0x13    /* Entered on reset, or execution of SVC instruction */
   .equ     Mode_MON,    0x16    /* Mode to swith between secure and non-secure state, or SMC call */
   .equ     Mode_ABT,    0x17    /* Entered as result of a data abort exception or prefetch abort exception */
   .equ     Mode_UNDEF,  0x1B    /* Entered as a result of an instruction-related error */
   .equ     Mode_SYS,    0x1F    /* Priveleged accesses  */
   .equ     Mode_MASK,   0x1F    /* Mask */
   .equ     IRQ_disable, 0x80    /* Mask to disable IRQ bit in CPSR */
   .equ     FIQ_disable, 0x40    /* Mask to disable FIQ bit in CPSR */

   /*
    * CR1 bits
    */ 
   .equ     CP15_CR_V, (1 << 13)
   
   /*
    * Stack fill value
    */
   .equ     STACK_FILL_VALUE, 0xAAAAAAAA 


/*=======================================================================*/
/*  Code Sections                                                        */
/*=======================================================================*/


   .section .isr_vectors, "ax"
   .code 32
   
/*************************************************************************/
/*  Exceptions vectors                                                   */
/*************************************************************************/
_start:
isr_vectors:
   ldr   pc, reset_addr
   ldr   pc, undef_addr
   ldr   pc, svc_addr
   ldr   pc, prefetch_addr
   ldr   pc, abort_addr
   ldr   pc, reserve_addr
   ldr   pc, irq_addr
   ldr   pc, fiq_addr
   
reset_addr:
   .word reset_handler
undef_addr:
   .word undef_handler   
svc_addr:
   .word svc_handler
prefetch_addr:
   .word prefetch_handler
abort_addr:
   .word abort_handler
reserve_addr:
   .word reserve_handler
irq_addr:
   .word irq_handler
fiq_addr:
   .word fiq_handler


  .section .init, "ax"
  .code 32
  .balign 4

/*************************************************************************/
/*  Reset handler and true entry point                                   */
/*************************************************************************/
reset_handler:

   /* Switch to SVC32 mode with disabled interrupts */
   mrs   r0, cpsr
   bic   r0, r0, #0x1F
   orr   r0, r0, #(Mode_SVC + IRQ_disable + FIQ_disable) /* 0xD3 */
   msr   cpsr_cxsf,r0
       
   /* Check which cpu is executing (just in-case) */
   mrc   p15, 0, r0, c0, c0, 5   /* Read arm affinity register */
   ands  r0, r0, #3              /* AND with 3 since only 2 cpus available*/
   beq   primary_cpu_init

   /* If here one of secondary cpu's reached this point */
secondary_cpu_loop:
   /* wfe */
   b     secondary_cpu_loop

   /* 
    * Primary cpu init must invalidate instruction cache left-on/used by BootROM 
    */
primary_cpu_init:

   /* Clear bit V of SCTRL Register */
   mrc   p15, 0, r0, c1, c0, 0
   bic   r0, #CP15_CR_V
   mcr   p15, 0, r0, c1, c0, 0

   /* Set VBAR */
   ldr   r0, =isr_vectors
   mcr   p15, 0, r0, c12, c0, 0    
   
   /* Set up stack pointer with symbol from linker script */
   LDR   sp, =__stack_end__
   
   mov   r0, #0

   /* Invalidate TLBs */
   mcr   p15, 0, r0, c8, c7, 0   

   /* Invalidate icache */
   mcr   p15, 0, r0, c7, c5, 0
   
   /* Invalidate branch predictor */
   mcr   p15, 0, r0, c7, c5, 6

   /* DSB instruction */
   mcr   p15, 0, r0, c7, c10, 4
   
   /* ISB instruction */
   mcr   p15, 0, r0, c7, c5, 4

   /* MMU and cache setup:  clear V, C, A, M bits */
   mrc   p15, 0, r0, c1, c0, 0
   bic   r0, r0, #0x00000007
   bic   r0, r0, #0x00002000

   /* Set A, Z, I bits */
   orr   r0, r0, #0x00000002  /* Set A, strict alignment fault checking enabled. */
   orr   r0, r0, #0x00000800  /* Set Z, Program flow prediction enabled. */
   orr   r0, r0, #0x00001000  /* Set I, Instruction caching enabled. */
   mcr   p15, 0, r0, c1, c0, 0
   nop
   nop
   nop
   nop
   

   /* 
    * Clear bss section (zero-init) 
    */
   ldr   r0, =__bss_start__
   ldr   r1, =__bss_end__
   mov   r2, #0
bss_loop:
   cmp   r0, r1      /* Compare addres */
   beq   bss_loop_end
   str   r2, [r0]    /* Clear location */
   add   r0, r0, #4  /* ptr += 4 */
   b     bss_loop    /* Loop until section end */
bss_loop_end:


   /*
    * Clear stack section, use fill value 0xAAAAAAAA
    */
   ldr   r0, =__stack_start__
   ldr   r1, =__stack_end__
   ldr   r2, =STACK_FILL_VALUE
stack_loop:
   cmp   r0, r1      /* Compare addres */
   beq   stack_loop_end
   str   r2, [r0]    /* Clear location */
   add   r0, r0, #4  /* ptr += 4 */
   b     stack_loop  /* Loop until section end */
stack_loop_end:


   /*
    * Setup stack pointer
    */
   mrs   r0, cpsr    /* Save original mode */

   /* Initialise the Undef mode stack */
   bic   r1, r0, #Mode_MASK
   orr   r1, r1, #Mode_UNDEF
   msr   cpsr_c, r1
   ldr   r13, =__und_stack_top__

   /* Initialise the Abort mode stack */
   bic   r1, r0, #Mode_MASK
   orr   r1, r1, #Mode_ABT
   msr   cpsr_c, r1
   ldr   r13, =__abt_stack_top__

   /* Initialise the IRQ mode stack */
   bic   r1, r0, #Mode_MASK
   orr   r1, r1, #Mode_IRQ
   msr   cpsr_c, r1
   ldr   r13, =__irq_stack_top__

   /* Initialise the FIQ mode stack */
   bic   r1, r0, #Mode_MASK
   orr   r1, r1, #Mode_FIQ
   msr   cpsr_c, r1
   ldr   r13, =__fiq_stack_top__

   /* Initialise the SVC mode stack */
   bic   r1, r0, #Mode_MASK
   orr   r1, r1, #Mode_SVC
   msr   cpsr_c, r1
   ldr   r13, =__svc_stack_top__

   /* Return to the original mode */
   msr   cpsr_c, r0


   /*
    * Enable FPU
    */
#if !defined(__SOFTFP__)
   /* Set the CPACR for access to CP10 and CP11 (the FPU coprocessors) */ 
   mrc   p15, 0, r0, c1, c0, 2
   ldr   r0, =(0xF << 20)
   mcr   p15, 0, r0, c1, c0, 2   
   
   /* Set the FPEXC EN bit to enable the FPU */
   mov   r3, #0x40000000 
   vmsr  fpexc, r3   
#endif

    
   /*
    * Branch to ARM main
    */
   bl    main

   
   /*
    * Return from main here
    */
inf_loop:
   wfi
   b     inf_loop


/*************************************************************************/
/*  Weak exceptions handler                                              */
/*************************************************************************/
undef_handler:
   b     undef_handler
   
svc_handler:
   b     svc_handler
   
prefetch_handler:
   b     prefetch_handler
   
abort_handler:
   b     abort_handler
   
reserve_handler:
   b     reserve_handler
   
fiq_handler:
   b     fiq_handler
   
   .weak undef_handler 
   .weak svc_handler
   .weak prefetch_handler 
   .weak abort_handler
   .weak reserve_handler
   .weak fiq_handler

/*************************************************************************/
/*  irq_handler                                                          */
/*************************************************************************/
   .global irq_handler
   .type irq_handler, function

/* http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.faqs/ka13552.html */
   
irq_handler:

   /* Save registers */
   sub   lr, lr, #4           /* Correct return address */
   push  {r0-r3, r12, lr}     /* Store AAPCS regs on to IRQ stack */

   /* Save FPU register */   
#if !defined(__SOFTFP__)
   vpush {d0-d15}             /* Save registers. */
   vmrs  r0, fpscr            /* Get status... */    
   push  {r0}                 /* ...and save status. */
#endif   
   
   /* The IRQ stack need a 8 byte alignment */
   mov   r0, sp                  
   and   r0, r0, #4           /* Ensure 8-byte stack alignment*/
   sub   sp, sp, r0           /* adjust stack as necessary */
   push  {r0}                 /* Store adjustment */
  
   /* Call the IRQHandler */
   blx   IRQHandler
     
   /* Restore the old IRQ stack alignment */
   pop   {r0}
   add   sp, sp, r0           /* Un-adjust stack */
   
   /* Restore FPU register */   
#if !defined(__SOFTFP__)
   pop   {r0}                 /* Get saved status... */
   vmsr  fpscr, r0            /* ...and restore status. */
   vpop  {d0-d15}             /* Restore registers. */
#endif
  
   /* Restore saved registers and return */
   pop   {r0-r3, r12, lr}     /* Restore AAPCS registers */
   movs  pc, lr               /* Return from interrupt */

/*** EOF ***/
